/*
 * Andrea Di Biagio
 * Politecnico di Milano, 2007
 * 
 * axe_debug.c
 * Formal Languages & Compilers Machine, 2007/2008
 * The ACSE software is distributed under the terms of the General Public License, version 2; use it at your own risk.
 * 
 */

#include "axe_debug.h"
#include "collections.h"
#include "reg_alloc_constants.h"
#include "cflow_constants.h"

static void printListOfVariables(t_list *variables, FILE *fout);
static void printBBlockInfos(t_basic_block *block, FILE *fout, int verbose);
static void printLiveIntervals(t_list *intervals, FILE *fout);
static void printBindings(int *bindings, int numVars, FILE *fout);

void printBindings(int *bindings, int numVars, FILE *fout)
{
   int counter;
   
   if (bindings == NULL)
      return;

   if (fout == NULL)
      return;

   /* initialize counter */
   counter = 0;
   fprintf(fout, "BINDINGS : \n");
   while(counter <= numVars)
   {
      if (bindings[counter] != RA_SPILL_REQUIRED)
      {
         fprintf(fout, "VAR T%d will be assigned to register R%d \n"
                  , counter, bindings[counter]);
      }
      else
      {
         fprintf(fout, "VAR T%d will be spilled \n", counter);
      }
      
      counter++;
   }
}

void printRegAllocInfos(t_reg_allocator *RA, FILE *fout)
{
   if (RA == NULL)
      return;
   if (fout == NULL)
      return;
   fprintf(fout, "\n\n*************************\n");
   fprintf(fout, "REGISTER ALLOCATION INFOS\n");
   fprintf(fout, "*************************\n");
   fprintf(fout, "AVAILABLE REGISTERS : %d \n", RA->regNum + 3);
   fprintf(fout, "USED VARIABLES : %d \n", RA->varNum);
   fprintf(fout, "-------------------------\n");
   printLiveIntervals(RA->live_intervals, fout);
   fprintf(fout, "-------------------------\n");
   printBindings(RA->bindings, RA->varNum, fout);
   fprintf(fout, "*************************\n\n");
}

void printLiveIntervals(t_list *intervals, FILE *fout)
{
   t_list *current_element;
   t_live_interval *interval;

   /* precondition */
   if (fout == NULL)
      return;

   fprintf(fout, "LIVE_INTERVALS:\n");

   /* retireve the first element of the list */
   current_element = intervals;
   while (current_element != NULL)
   {
      interval = (t_live_interval *) LDATA(current_element);

      fprintf(fout, "\tLIVE_INTERVAL of T%d : [%d, %d] \n"
            , interval->varID, interval->startPoint, interval->endPoint);
      
      /* retrieve the next element in the list of intervals */
      current_element = LNEXT(current_element);
   }
}

void printBBlockInfos(t_basic_block *block, FILE *fout, int verbose)
{
   t_list *current_element;
   t_cflow_Node *current_node;
   int count;
   
   /* preconditions */
   if (block == NULL)
      return;
   if (fout == NULL)
      return;

   fprintf(fout,"NUMBER OF PREDECESSORS : %d \n"
         , getLength(block->pred) );
   fprintf(fout,"NUMBER OF SUCCESSORS : %d \n"
         , getLength(block->succ) );
   fprintf(fout,"NUMBER OF INSTRUCTIONS : %d \n"
         , getLength(block->nodes) );

   count = 1;
   current_element = block->nodes;
   while(current_element != NULL)
   {
      current_node = (t_cflow_Node *) LDATA(current_element);
      fprintf(fout,"\t%d.  ", count);
      debug_printInstruction(current_node->instr, fout);
      if (verbose != 0)
      {
         if (current_node->def != NULL)
            fprintf(fout, "\n\t\t\tDEF = [R%d]", (current_node->def)->ID);
         if (current_node->uses[0] != NULL)
         {
            fprintf(fout, "\n\t\t\tUSES = [R%d", ((current_node->uses)[0])->ID);
            if (current_node->uses[1] != NULL)
               fprintf(fout, ", R%d", ((current_node->uses)[1])->ID);
            if (current_node->uses[2] != NULL)
               fprintf(fout, ", R%d", ((current_node->uses)[2])->ID);
            fprintf(fout, "]");
         }
         fprintf(fout, "\n\t\t\tLIVE IN = [");
         printListOfVariables(current_node->in, fout);
         fprintf(fout, "]");
         fprintf(fout, "\n\t\t\tLIVE OUT = [");
         printListOfVariables(current_node->out, fout);
         fprintf(fout, "]");
      }
      
      fprintf(fout, "\n");
      count++;
      current_element = LNEXT(current_element);
   }
}

void printListOfVariables(t_list *variables, FILE *fout)
{
   t_list *current_element;
   t_cflow_var *current_variable;
   
   if (variables == NULL)
      return;
   if (fout == NULL)
      return;

   current_element = variables;
   while(current_element != NULL)
   {
      current_variable = (t_cflow_var *) LDATA(current_element);
      fprintf(fout, "R%d", current_variable->ID);
      if (LNEXT(current_element) != NULL)
         fprintf(fout, ", ");
      
      current_element = LNEXT(current_element);
   }
}

void printGraphInfos(t_cflow_Graph *graph, FILE *fout, int verbose)
{
   int counter;
   t_list *current_element;
   t_basic_block *current_bblock;
   
   /* preconditions */
   if (graph == NULL)
      return;
   if (fout == NULL)
      return;

   /* initialization of the local variables */
   counter = 1;
   
   fprintf(fout,"NOTE : Temporary registers are considered as\n"
                "       variables of the intermediate language. \n");
#if CFLOW_ALWAYS_LIVEIN_R0 == (1)
   fprintf(fout,"       Variable \'R0\' (that refers to the \n"
                "       physical register \'RO\') is always \n"
                "       considered LIVE-IN for each node of \n"
                "       a basic block. \n"
                "       Thus, in the following control flow graph, \n"
                "       \'R0\' will never appear as LIVE-IN or LIVE-OUT\n"
                "       variable for a statement.\n\n"
                "       If you want to consider \'R0\' as\n"
                "       a normal variable, you have to set\n"
                "       to 0 the value of the macro CFLOW_ALWAYS_LIVEIN_R0\n"
                "       defined in \"cflow_constants.h\".\n\n");
#endif
   fprintf(fout,"\n");
   fprintf(fout,"**************************\n");
   fprintf(fout,"     CONTROL FLOW GRAPH   \n");
   fprintf(fout,"**************************\n");
   fprintf(fout,"NUMBER OF BASIC BLOCKS : %d \n"
         , getLength(graph->blocks));
   fprintf(fout,"NUMBER OF USED VARIABLES : %d \n"
         , getLength(graph->cflow_variables));
   fprintf(fout,"--------------------------\n");
   fprintf(fout,"START BASIC BLOCK INFOS.  \n");
   fprintf(fout,"--------------------------\n");

   /* initialize `current_block' */
   current_element = graph->blocks;
   while(current_element != NULL)
   {
      current_bblock = (t_basic_block *) LDATA(current_element);
      fprintf(fout,"[BLOCK %d] \n", counter);
      printBBlockInfos(current_bblock, fout, verbose);
      if (LNEXT(current_element) != NULL)
         fprintf(fout,"--------------------------\n");
      else
         fprintf(fout,"**************************\n");

      counter++;
      current_element = LNEXT(current_element);
   }
   
   fprintf(fout,"\n\n");
}


void debug_printInstruction(t_axe_instruction *instr, FILE *fout)
{
   /* preconditions */
   if (fout == NULL)
      return;
   
   if (instr == NULL)
   {
      fprintf(fout, "[NULL] \n");
      return;
   }

   if (instr->labelID != NULL)
      fprintf(fout, "L%d\t", (instr->labelID)->labelID);
   else
      fprintf(fout, "\t");
   
   switch(instr->opcode)
   {
      case ADD : fprintf(fout, "jit_addr_i"); break;
      case SUB : fprintf(fout, "jit_subr_i "); break;
      case ANDL : fprintf(fout, "jit_andr_i "); break;
      case ORL : fprintf(fout, "jit_orr_i "); break;
      case EORL : fprintf(fout, "jit_xor_i "); break;
      case JIT_ADDCR_I : fprintf(fout, "jit_addcr_i "); break;
      case JIT_ADDXR_I : fprintf(fout, "jit_addxr_i "); break;
      case JIT_SUBCR_I : fprintf(fout, "jit_subcr_i "); break;
      case MUL : fprintf(fout, "jit_mulr_i"); break;
      case DIV : fprintf(fout, "jit_divr_i "); break;
      case SHL : fprintf(fout, "jit_lshr_i "); break;
      case SHR : fprintf(fout, "jit_rshr_i "); break;
      case JIT_SUBXR_I : fprintf(fout, "jit_subxr_i "); break;
      case JIT_HMULR_I : fprintf(fout, "jit_hmulr_i "); break;
      case JIT_MODR_I : fprintf(fout, "jit_modr_i "); break;
      case JIT_RSBR_I : fprintf(fout, "jit_rsbr_i "); break;
      case ADDI : fprintf(fout, "jit_addi_i "); break;
      case SUBI : fprintf(fout, "jit_subi_i "); break;
      case ANDLI : fprintf(fout, "jit_andi_i "); break;
      case ORLI : fprintf(fout, "jit_ori_i "); break;
      case EORLI : fprintf(fout, "jit_xori_i "); break;
      case JIT_ADDCI_I : fprintf(fout, "jit_addci_i "); break;
      case JIT_ADDXI_I : fprintf(fout, "jit_addxi_i "); break;
      case JIT_SUBCI_I : fprintf(fout, "jit_subci_i "); break;
      case MULI : fprintf(fout, "jit_muli_i"); break;
      case DIVI : fprintf(fout, "jit_divi_i"); break;
      case SHLI : fprintf(fout, "jit_lshi_i"); break;
      case SHRI : fprintf(fout, "jit_rshi_i"); break;
      case JIT_SUBXI_I : fprintf(fout, "jit_subxi_i "); break;
      case JIT_HMULI_I : fprintf(fout, "jit_hmuli_i "); break;
      case JIT_MODI_I : fprintf(fout, "jit_modi_i "); break;
      case JIT_RSBI_I : fprintf(fout, "jit_rsbi_i "); break;
      case NOP : fprintf(fout, "jit_nop "); break;
      case HALT : fprintf(fout, "jit_ret "); break;
      /*case SEQ : fprintf(fout, "SEQ "); break;
      case SGE : fprintf(fout, "SGE "); break;
      case SGT : fprintf(fout, "SGT "); break;
      case SLE : fprintf(fout, "SLE "); break;
      case SLT : fprintf(fout, "SLT "); break;
      case SNE : fprintf(fout, "SNE "); break;*/
      case BT : fprintf(fout, "jit_jmpi "); break;
      case BF : fprintf(fout, "jit_jmpi "); break;
      /*case BHI : fprintf(fout, "BHI "); break;
      case BLS : fprintf(fout, "BLS "); break;
      case BCC : fprintf(fout, "BCC "); break;
      case BCS : fprintf(fout, "BCS "); break;
      case BNE : fprintf(fout, "BNE "); break;
      case BEQ : fprintf(fout, "BEQ "); break;
      case BVC : fprintf(fout, "BVC "); break;
      case BVS : fprintf(fout, "BVS "); break;
      case BPL : fprintf(fout, "BPL "); break;
      case BMI : fprintf(fout, "BMI "); break;
      case BGE : fprintf(fout, "BGE "); break;
      case BLT : fprintf(fout, "BLT "); break;
      case BGT : fprintf(fout, "BGT "); break;
      case BLE : fprintf(fout, "BLE "); break;*/
      case LOAD : fprintf(fout, "jit_ldi_i "); break;
      case STORE : fprintf(fout, "jit_str_i "); break;
      case AXE_READ : fprintf(fout, "scanf "); break;
      case AXE_WRITE : fprintf(fout, "printf "); break;
      case INVALID_OPCODE : fprintf(fout, "[INVALID] ");
   }

   if (instr->reg_1 != NULL)
   {
      if (!(instr->reg_1)->indirect)
         fprintf(fout, "R%d ", (instr->reg_1)->ID);
      else
         fprintf(fout, "(R%d) ", (instr->reg_1)->ID);
   }
   if (instr->reg_2 != NULL)
   {
      if (!(instr->reg_2)->indirect)
         fprintf(fout, "R%d ", (instr->reg_2)->ID);
      else
         fprintf(fout, "(R%d) ", (instr->reg_2)->ID);
      if (instr->reg_3 != NULL)
      {
         if (!(instr->reg_3)->indirect)
            fprintf(fout, "R%d ", (instr->reg_3)->ID);
         else
            fprintf(fout, "(R%d) ", (instr->reg_3)->ID);
      }
      else
         fprintf(fout, "#%d ", instr->immediate);
   }
   
   if (instr->address != NULL)
   {
      if ((instr->address)->type == LABEL_TYPE)
         fprintf(fout, "L%d ", ((instr->address)->labelID)->labelID);
      else
         fprintf(fout, "%d ", (instr->address)->addr);
   }
}

char * dataTypeToString(int codedType)
{
   switch (codedType)
   {
      case INTEGER_TYPE : return "INTEGER";
      default : return "<INVALID_TYPE>";
   }
}
