/*
 * Andrea Di Biagio
 * Politecnico di Milano, 2007
 * 
 * axe_expressions.c
 * Formal Languages & Compilers Machine, 2007/2008
 * The ACSE software is distributed under the terms of the General Public License, version 2; use it at your own risk.
 * 
 */

#include "axe_expressions.h"
#include "axe_gencode.h"
#include "axe_errors.h"


static t_axe_expression perform_bin_numeric_op_Imm
         (int val1, int val2, int binop);
         
static t_axe_expression perform_bin_comparison_Imm
         (int val1, int val2, int condition);


t_axe_expression perform_bin_numeric_op (t_program_infos *program
         , t_axe_expression exp1, t_axe_expression exp2, int binop)
{
   int output_register;
   t_axe_expression result;

   /* initialize result */
   result.expression_type = INVALID_EXPRESSION;

   /* we have to test if one (or both) of
   * the two operands is an immediate value */
   if (  (exp2.expression_type == IMMEDIATE)
         && (exp1.expression_type == IMMEDIATE) )
   {
      return perform_bin_numeric_op_Imm(exp1.value, exp2.value, binop);
   }
   
   /* at first we have to ask for a free register
   * where to store the result of the operation. */
   output_register = getNewRegister(program);

   if (exp2.expression_type == IMMEDIATE)
   {
      /* we have to produce an instruction */
      switch(binop)
      {
         case ADD : gen_addi_instruction (program, output_register
                             , exp1.value, exp2.value); break;
         case ANDL : gen_andli_instruction (program, output_register
                              , exp1.value, exp2.value); break;
         case ORL  : gen_orli_instruction (program, output_register
                             , exp1.value, exp2.value); break;
         case SUB : gen_subi_instruction (program, output_register
                             , exp1.value, exp2.value); break;
         case MUL : gen_muli_instruction (program, output_register
                             , exp1.value, exp2.value); break;
         case SHL : gen_shli_instruction (program, output_register
                             , exp1.value, exp2.value); break;
         case SHR : gen_shri_instruction (program, output_register
                             , exp1.value, exp2.value); break;
         case DIV :
               if (exp2.value == 0){
                  printWarningMessage(WARN_DIVISION_BY_ZERO);
               }
               gen_divi_instruction (program, output_register
                        , exp1.value, exp2.value);
               break;
         default :
               notifyError(AXE_INVALID_EXPRESSION);
      }
   }
   else if (exp1.expression_type == IMMEDIATE)
   {
      int other_reg;

      /* we have to produce an instruction */
      switch(binop)
      {
         case ADD :  gen_addi_instruction (program, output_register
                              , exp2.value, exp1.value); break;
         case ANDL :  gen_andli_instruction (program, output_register
                              , exp2.value, exp1.value); break;
         case ORL  :  gen_orli_instruction (program, output_register
                              , exp2.value, exp1.value); break;
         case SUB :
                  gen_subi_instruction (program, output_register
                           , exp2.value, exp1.value);

                  /* we have to produce a NEG instruction */
                  gen_neg_instruction (program, output_register
                           , output_register, CG_DIRECT_ALL);
                  break;
         case MUL :  gen_muli_instruction (program, output_register
                              , exp2.value, exp1.value); break;
         case DIV :
                  /* we have to load into a register the immediate value */
                  other_reg = getNewRegister(program);

                  /* In order to load the immediate inside a new
                   * register we have to insert an ADDI instr. */
                  gen_addi_instruction (program, other_reg
                           , REG_0, exp1.value);

                  /* we have to produce a DIV instruction */
                  gen_div_instruction (program, output_register
                           , other_reg, exp2.value, CG_DIRECT_ALL);
                  break;
         case SHL :
                  /* we have to load into a register the immediate value */
                  other_reg = getNewRegister(program);

                  /* In order to load the immediate inside a new
                   * register we have to insert an ADDI instr. */
                  gen_addi_instruction (program, other_reg
                           , REG_0, exp1.value);

                  /* we have to produce a SHL instruction */
                  gen_shl_instruction (program, output_register
                           , other_reg, exp2.value, CG_DIRECT_ALL);
                  break;
         case SHR :
                  /* we have to load into a register the immediate value */
                  other_reg = getNewRegister(program);

                  /* In order to load the immediate inside a new
                   * register we have to insert an ADDI instr. */
                  gen_addi_instruction (program, other_reg
                           , REG_0, exp1.value);

                  /* we have to produce a SHR instruction */
                  gen_shr_instruction (program, output_register
                           , other_reg, exp2.value, CG_DIRECT_ALL);
                  break;
         default :
                  notifyError(AXE_INVALID_EXPRESSION);
      }
   }
   else
   {
      /* we have to produce an instruction */
      switch(binop)
      {
         case ADD :  gen_add_instruction (program, output_register
                              , exp1.value, exp2.value, CG_DIRECT_ALL);
                     break;
         case ANDL :  gen_andl_instruction (program, output_register
                              , exp1.value, exp2.value, CG_DIRECT_ALL);
                     break;
         case ORL :  gen_orl_instruction (program, output_register
                              , exp1.value, exp2.value, CG_DIRECT_ALL);
                     break;
         case SUB :  gen_sub_instruction (program, output_register
                              , exp1.value, exp2.value, CG_DIRECT_ALL);
                     break;
         case MUL :  gen_mul_instruction (program, output_register
                              , exp1.value, exp2.value, CG_DIRECT_ALL);
                     break;
         case DIV :  gen_div_instruction (program, output_register
                              , exp1.value, exp2.value, CG_DIRECT_ALL);
                     break;
         case SHL :  gen_shl_instruction (program, output_register
                              , exp1.value, exp2.value, CG_DIRECT_ALL);
                     break;
         case SHR :  gen_shl_instruction (program, output_register
                              , exp1.value, exp2.value, CG_DIRECT_ALL);
                     break;
         default :
                     notifyError(AXE_INVALID_EXPRESSION);
      }
   }
         
   /* assign a value to result */
   return create_expression (output_register, REGISTER);
}

t_axe_expression perform_bin_numeric_op_Imm
         (int val1, int val2, int binop)
{
   switch(binop)
   {
      case ADD : return create_expression ((val1 + val2), IMMEDIATE);
      case ANDL : return create_expression ((val1 & val2), IMMEDIATE);
      case ORL  : return create_expression ((val1 | val2), IMMEDIATE);
      case SUB : return create_expression ((val1 - val2), IMMEDIATE);
      case MUL : return create_expression ((val1 * val2), IMMEDIATE);
      case SHL : return create_expression ((val1 << val2), IMMEDIATE);
      case SHR : return create_expression ((val1 >> val2), IMMEDIATE);
      case DIV :
         if (val2 == 0){
            printWarningMessage(WARN_DIVISION_BY_ZERO);
         }
         return create_expression ((val1 / val2), IMMEDIATE);
      default :
         notifyError(AXE_INVALID_EXPRESSION);
   }
   
   return create_expression (0, INVALID_EXPRESSION);
}

t_axe_expression perform_bin_comparison_Imm
         (int val1, int val2, int condition)
{
   switch(condition)
   {
      case _LT_ : return create_expression ((val1 < val2), IMMEDIATE);
      case _GT_ : return create_expression ((val1 > val2), IMMEDIATE);
      case _EQ_  : return create_expression ((val1 == val2), IMMEDIATE);
      case _NOTEQ_ : return create_expression ((val1 != val2), IMMEDIATE);
      case _LTEQ_ : return create_expression ((val1 <= val2), IMMEDIATE);
      case _GTEQ_ : return create_expression ((val1 >= val2), IMMEDIATE);
      default :
         notifyError(AXE_INVALID_EXPRESSION);
   }

   return create_expression (0, INVALID_EXPRESSION);
}

t_axe_expression perform_binary_comparison (t_program_infos *program
         , t_axe_expression exp1, t_axe_expression exp2, int condition)
{
   t_axe_expression result;
   int output_register;
   
   /* initialize result */
   result.expression_type = INVALID_EXPRESSION;

   /* we have to test if one (or both) of
   * the two operands is an immediate value */
   if (  (exp2.expression_type == IMMEDIATE)
         && (exp1.expression_type == IMMEDIATE) )
   {
      return perform_bin_comparison_Imm
                  (exp1.value, exp2.value, condition);
   }
                     
   /* at first we have to ask for a free register
   * where to store the result of the comparison. */
   output_register = getNewRegister(program);

   if (exp2.expression_type == IMMEDIATE)
   {
   	switch(condition)
   	{
      		case _LT_ : gen_jit_lti_i_instruction (program, output_register, exp1.value, exp2.value); break;
      		case _GT_ : gen_jit_gti_i_instruction (program, output_register, exp1.value, exp2.value ); break;
      		case _EQ_  : gen_jit_eqi_i_instruction (program, output_register, exp1.value, exp2.value); break;
      		case _NOTEQ_ : gen_jit_nei_i_instruction (program, output_register, exp1.value, exp2.value); break;
      		case _LTEQ_ : gen_jit_lei_i_instruction (program, output_register, exp1.value, exp2.value); break;
      		case _GTEQ_ : gen_jit_gei_i_instruction (program, output_register, exp1.value, exp2.value); break;
      		default :
         		notifyError(AXE_INVALID_EXPRESSION);
   	}

   }
   else if (exp1.expression_type == IMMEDIATE)
   {
   	switch(condition)
   	{
      		case _LT_ : gen_jit_gti_i_instruction (program, output_register, exp2.value, exp1.value); break;
      		case _GT_ : gen_jit_lti_i_instruction (program, output_register, exp2.value, exp1.value); break;
      		case _EQ_  : gen_jit_eqi_i_instruction (program, output_register, exp2.value, exp1.value); break;
      		case _NOTEQ_ : gen_jit_nei_i_instruction (program, output_register, exp2.value, exp1.value); break;
      		case _LTEQ_ : gen_jit_gei_i_instruction (program, output_register, exp2.value, exp1.value); break;
      		case _GTEQ_ : gen_jit_lei_i_instruction (program, output_register, exp2.value, exp1.value); break;
      		default :
         		notifyError(AXE_INVALID_EXPRESSION);
   	}	
	
   }
   else
   {
   	switch(condition)
   	{
      		case _LT_ : gen_jit_gtr_i_instruction (program, output_register, exp1.value, exp2.value,0); break;
      		case _GT_ : gen_jit_ltr_i_instruction (program, output_register, exp1.value, exp2.value,0); break;
      		case _EQ_  : gen_jit_eqr_i_instruction (program, output_register, exp1.value, exp2.value,0); break;
      		case _NOTEQ_ : gen_jit_ner_i_instruction (program, output_register, exp1.value, exp2.value,0); break;
      		case _LTEQ_ : gen_jit_ger_i_instruction (program, output_register, exp1.value, exp2.value,0); break;
      		case _GTEQ_ : gen_jit_ler_i_instruction (program, output_register, exp1.value, exp2.value,0); break;
      		default :
         		notifyError(AXE_INVALID_EXPRESSION);
   	}
   }

   /* return the new expression */
   return create_expression (output_register, REGISTER);
}

/*
t_axe_expression perform_branch (t_program_infos *program
         , t_axe_expression exp1, t_axe_expression exp2, int condition, t_axe_label *label)
{
   t_axe_expression result;
   //int output_register;
   
   /* initialize result 
   result.expression_type = INVALID_EXPRESSION;

   /* we have to test if one (or both) of
   * the two operands is an immediate value 
   if (  (exp2.expression_type == IMMEDIATE)
         && (exp1.expression_type == IMMEDIATE) )
   {
      return perform_bin_comparison_Imm
                  (exp1.value, exp2.value, condition);
   }
                     
   /* at first we have to ask for a free register
   * where to store the result of the comparison. 
   //output_register = getNewRegister(program);

   if (exp2.expression_type == IMMEDIATE)
   {
   	switch(condition)
   	{
      		case _LT_ : gen_jit_blti_i_instruction (program, label, exp1.value, exp2.value, 0); break;
      		case _GT_ : gen_jit_bgti_i_instruction (program, label, exp1.value, exp2.value,0); break;
      		case _EQ_  : gen_jit_beqi_i_instruction (program, ,label, exp1.value, exp2.value,0); break;
      		case _NOTEQ_ : gen_jit_bnei_i_instruction (program, label, exp1.value, exp2.value,0); break;
      		case _LTEQ_ : gen_jit_blei_i_instruction (program, label, exp1.value, exp2.value,0); break;
      		case _GTEQ_ : gen_jit_bgei_i_instruction (program, label, exp1.value, exp2.value,0); break;
      		default :
         		notifyError(AXE_INVALID_EXPRESSION);
   	}

   }
   else if (exp1.expression_type == IMMEDIATE)
   {
   	switch(condition)
   	{
      		case _LT_ : gen_jit_bgti_i_instruction (program, label, exp2.value, exp1.value,0); break;
      		case _GT_ : gen_jit_blti_i_instruction (program, label, exp2.value, exp1.value,0); break;
      		case _EQ_  : gen_jit_beqi_i_instruction (program, label, exp2.value, exp1.value,0); break;
      		case _NOTEQ_ : gen_jit_bnei_i_instruction (program, label, exp2.value, exp1.value,0); break;
      		case _LTEQ_ : gen_jit_bgei_i_instruction (program, label, exp2.value, exp1.value,0); break;
      		case _GTEQ_ : gen_jit_blei_i_instruction (program, label, exp2.value, exp1.value,0); break;
      		default :
         		notifyError(AXE_INVALID_EXPRESSION);
   	}	
	
   }
   else
   {
   	switch(condition)
   	{
      		case _LT_ : gen_jit_bgtr_i_instruction (program, label, exp1.value, exp2.value, 0); break;
      		case _GT_ : gen_jit_bltr_i_instruction (program, label, exp1.value, exp2.value, 0); break;
      		case _EQ_  : gen_jit_beqr_i_instruction (program, label, exp1.value, exp2.value, 0); break;
      		case _NOTEQ_ : gen_jit_bner_i_instruction (program, label, exp1.value, exp2.value, 0); break;
      		case _LTEQ_ : gen_jit_bger_i_instruction (program, label, exp1.value, exp2.value, 0); break;
      		case _GTEQ_ : gen_jit_bler_i_instruction (program, label, exp1.value, exp2.value, 0); break;
      		default :
         		notifyError(AXE_INVALID_EXPRESSION);
   	}
   }

   /* return the new expression 
   return create_expression (0, IVALID_EXPRESSION);
}*/
