/*
 * Andrea Di Biagio
 * Politecnico di Milano, 2007
 * 
 * axe_gencode.c
 * Formal Languages & Compilers Machine, 2007/2008
 * 
 */

#include "axe_gencode.h"
#include "axe_errors.h"


static t_axe_instruction * gen_unary_instruction (t_program_infos *program
      , int opcode, int r_dest, t_axe_label *label, int addr);
static t_axe_instruction * gen_unary_rw_instruction (t_program_infos *program
      , int opcode, int r_dest);
static t_axe_instruction * gen_unary_immediate_instruction(t_program_infos *program
      , int opcode, int r_dest, int immediate);
static t_axe_instruction * gen_binary_instruction (t_program_infos *program
      , int opcode, int r_dest, int r_source1, int immediate);
static t_axe_instruction * gen_ternary_instruction (t_program_infos *program
      , int opcode, int r_dest, int r_source1, int r_source2, int flags);
static t_axe_instruction * gen_bin_jump_instruction (t_program_infos *program
      , int opcode, int r_source1, int r_source2, t_axe_label *label, int addr);
static t_axe_instruction * gen_unary_immediate_jump_instruction (t_program_infos *program
      , int opcode, int r_source1,int immediate, t_axe_label *label, int addr);
static t_axe_instruction * gen_load_store_instruction (t_program_infos *program
		, int opcode, int r_dest, int r_source1,int immediate, t_axe_label *label, int addr);
static t_axe_instruction * gen_jump_instruction (t_program_infos *program
      , int opcode, t_axe_label *label, int addr);



/*----------------------------------------------------
 *                  TERNARY INSTRUCTIONS
 *---------------------------------------------------*/
/* ADDR Addition*/
t_axe_instruction * gen_add_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
   return gen_ternary_instruction (program, ADD, r_dest, r_source1, r_source2, flags);
}

/* ADDXR Addition which checks the carry*/
t_axe_instruction * gen_jit_addxr_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
	return gen_ternary_instruction (program, JIT_ADDXR_I, r_dest, r_source1, r_source2, flags);
}

/* ADDCR Addition with carry*/
t_axe_instruction * gen_jit_addcr_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
	return gen_ternary_instruction (program, JIT_ADDCR_I, r_dest, r_source1, r_source2,flags);
}
/* SUBR Subtraction */
t_axe_instruction * gen_sub_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
   return gen_ternary_instruction(program, SUB, r_dest, r_source1, r_source2, flags);
}

/* SUBXR Subtraction which cheks the carry*/
t_axe_instruction * gen_jit_subxr_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2,int flags)
{
	return gen_ternary_instruction (program, JIT_SUBXR_I, r_dest, r_source1, r_source2, flags);
}

/* SUBCR Subtraction with carry*/
t_axe_instruction * gen_jit_subcr_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
	return gen_ternary_instruction (program, JIT_SUBCR_I, r_dest, r_source1, r_source2, flags);
}

/* RSBR Reverse substraction*/
t_axe_instruction * gen_jit_rsbr_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
	return gen_ternary_instruction (program, JIT_RSBR_I, r_dest, r_source1, r_source2, flags);
}

/* ANDR Logical and*/
t_axe_instruction * gen_andl_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
   return gen_ternary_instruction (program, ANDL, r_dest, r_source1, r_source2, flags);
}

/* ORR Logical or*/
t_axe_instruction * gen_orl_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
   return gen_ternary_instruction (program, ORL, r_dest, r_source1, r_source2, flags);
}
/* XORR Exclusive or*/
t_axe_instruction * gen_eorl_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
   return gen_ternary_instruction (program, EORL, r_dest, r_source1, r_source2, flags);
}

/* MULR Multiplication */
t_axe_instruction * gen_mul_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
   return gen_ternary_instruction (program, MUL, r_dest, r_source1, r_source2, flags);
}

/* HMULR Multiplication of high bits*/
t_axe_instruction * gen_jit_hmulr_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
	return gen_ternary_instruction (program, JIT_HMULR_I, r_dest, r_source1, r_source2, flags);
}

/* DIVR Division*/
t_axe_instruction * gen_div_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
   return gen_ternary_instruction (program, DIV, r_dest, r_source1, r_source2, flags);
}

/* MODR Modulo*/
t_axe_instruction * gen_jit_modr_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
	return gen_ternary_instruction (program, JIT_MODR_I, r_dest, r_source1, r_source2, flags);
}

/* LSHR Shift left*/
t_axe_instruction * gen_shl_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
   return gen_ternary_instruction (program, SHL, r_dest, r_source1, r_source2, flags);
}

/* RSHR Shift right*/
t_axe_instruction * gen_shr_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
   return gen_ternary_instruction (program, SHL, r_dest, r_source1, r_source2, flags);
}

/* LTR Lower than*/
t_axe_instruction * gen_jit_ltr_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
	return gen_ternary_instruction (program, JIT_LTR_I, r_dest, r_source1, r_source2, flags);
}

/* LER Lower or equal*/
t_axe_instruction * gen_jit_ler_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
	return gen_ternary_instruction (program, JIT_LER_I, r_dest, r_source1, r_source2, flags);
}

/* GTR Greater than*/
t_axe_instruction * gen_jit_gtr_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
	return gen_ternary_instruction (program, JIT_GTR_I, r_dest, r_source1, r_source2, flags);
}

/* GER Greater or equal*/
t_axe_instruction * gen_jit_ger_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
	return gen_ternary_instruction (program, JIT_GER_I, r_dest, r_source1, r_source2, flags);
}

/* EQR Equal*/
t_axe_instruction * gen_jit_eqr_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
	return gen_ternary_instruction (program, JIT_EQR_I, r_dest, r_source1, r_source2, flags);
}

/* NER Not equal*/
t_axe_instruction * gen_jit_ner_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int r_source2, int flags)
{
	return gen_ternary_instruction (program, JIT_NER_I, r_dest, r_source1, r_source2, flags);
}

/*NEGR Negative*/
t_axe_instruction * gen_neg_instruction (t_program_infos *program
      , int r_dest, int r_source, int flags)
{
   return gen_ternary_instruction(program, SUB, r_dest, REG_0, r_source, flags);
}


/*-------------------------------------------------------
  *            BINARY  INSTRUCTIONS
  *------------------------------------------------------*/

/* ADDI Addition immediate*/
t_axe_instruction * gen_addi_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate)
{
   return gen_binary_instruction (program, ADDI, r_dest, r_source1, immediate);
}

/* ADDXI Addition immediate which checks the carry*/
t_axe_instruction * gen_jit_addxi_i_instruction(t_program_infos *program
	, int r_dest, int r_source1, int immediate)
{
	return gen_binary_instruction (program, JIT_ADDXI_I, r_dest, r_source1, immediate);
}

/* ADDCI Addition immediate with carry*/
t_axe_instruction * gen_jit_addci_i_instruction(t_program_infos *program
	, int r_dest, int r_source1, int immediate)
{
	return gen_binary_instruction (program, JIT_ADDCI_I, r_dest, r_source1, immediate);
}

/* SUBI Substraction immediate*/
t_axe_instruction * gen_subi_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate)
{
   return gen_binary_instruction (program, SUBI, r_dest, r_source1, immediate);
}

/* SUBXI Substraction immediate which checks carry*/
t_axe_instruction * gen_jit_subxi_i_instruction(t_program_infos *program
	, int r_dest, int r_source1, int immediate)
{
	return gen_binary_instruction (program, JIT_SUBXI_I, r_dest, r_source1, immediate);
}

/* SUBCI Substraction immediate with carry*/
t_axe_instruction * gen_jit_subci_i_instruction(t_program_infos *program
	, int r_dest, int r_source1, int immediate)
{
	return gen_binary_instruction (program, JIT_SUBCI_I, r_dest, r_source1, immediate);
}

/* RSBI Reverse substraction immediate*/
t_axe_instruction * gen_jit_rsbi_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int immediate)
{
	return gen_binary_instruction (program, JIT_RSBI_I, r_dest, r_source1, immediate);
}

/* ANDI Logical and immediate*/
t_axe_instruction * gen_andli_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate)
{
   return gen_binary_instruction(program, ANDLI, r_dest, r_source1, immediate);
}

/*ORI Logical or immediate*/
t_axe_instruction * gen_orli_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate)
{
   return gen_binary_instruction(program, ORLI, r_dest, r_source1, immediate);
}

/*XORI Exclusive or immediate*/
t_axe_instruction * gen_eorli_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate)
{
   return gen_binary_instruction(program, EORLI, r_dest, r_source1, immediate);
}

/* MULI Multiplication immediate*/
t_axe_instruction * gen_muli_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate)
{
   return gen_binary_instruction(program, MULI, r_dest, r_source1, immediate);
}

/* HMULI Multiplication of high bits immediate*/
t_axe_instruction * gen_jit_hmuli_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int immediate)
{
	return gen_binary_instruction (program, JIT_HMULI_I, r_dest, r_source1, immediate);
}

/*DIVI Division immediate*/
t_axe_instruction * gen_divi_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate)
{
   return gen_binary_instruction(program, DIVI, r_dest, r_source1, immediate);
}

/* MODI Modulo immediate*/
t_axe_instruction * gen_jit_modi_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int immediate)
{
	return gen_binary_instruction (program, JIT_MODI_I, r_dest, r_source1, immediate);
}

/*LSHI Shift left immediate*/
t_axe_instruction * gen_shli_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate)
{
   return gen_binary_instruction(program, SHLI, r_dest, r_source1, immediate);
}

/*RSHI Shift right immediate*/
t_axe_instruction * gen_shri_instruction
      (t_program_infos *program, int r_dest, int r_source1, int immediate)
{
   return gen_binary_instruction(program, SHRI, r_dest, r_source1, immediate);
}

/* LTI Lower than immediate*/
t_axe_instruction * gen_jit_lti_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int immediate)
{
	return gen_binary_instruction (program, JIT_LTI_I, r_dest, r_source1, immediate);
}

/* LEI Lower or equal immediate*/
t_axe_instruction * gen_jit_lei_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int immediate)
{
	return gen_binary_instruction (program, JIT_LEI_I, r_dest, r_source1, immediate);
}

/* GTI Greater than immediate*/
t_axe_instruction * gen_jit_gti_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int immediate)
{
	return gen_binary_instruction (program, JIT_GTI_I, r_dest, r_source1, immediate);
}

/* GEI Greater or equal immediate*/
t_axe_instruction * gen_jit_gei_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int immediate)
{
	return gen_binary_instruction (program, JIT_GER_I, r_dest, r_source1, immediate);
}

/* EQI Equal immediate*/
t_axe_instruction * gen_jit_eqi_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int immediate)
{
	return gen_binary_instruction (program, JIT_EQI_I, r_dest, r_source1, immediate);
}

/* NEI Not equal immediate*/
t_axe_instruction * gen_jit_nei_i_instruction (t_program_infos *program
      , int r_dest, int r_source1, int immediate)
{
	return gen_binary_instruction (program, JIT_NEI_I, r_dest, r_source1, immediate);
}

/* NOTR*/
t_axe_instruction * gen_notl_instruction
      (t_program_infos *program, int r_dest, int r_source1)
{
   return gen_binary_instruction(program, NOTL, r_dest, r_source1, 0);
}
/*-------------------------------------------------------
  *            READ/WRITE  INSTRUCTIONS
  *------------------------------------------------------*/

t_axe_instruction * gen_read_instruction
               (t_program_infos *program, int r_dest)
{
   return gen_unary_rw_instruction(program, AXE_READ, r_dest);
}

t_axe_instruction * gen_write_instruction
               (t_program_infos *program, int r_dest)
{
   return gen_unary_rw_instruction(program, AXE_WRITE, r_dest);
}

/*-------------------------------------------------------
  *           LOAD/STORE  INSTRUCTIONS
  *------------------------------------------------------*/

t_axe_instruction * gen_load_instruction
      (t_program_infos *program, int r_dest, t_axe_label *label, int address)
{
   return gen_unary_instruction(program, LOAD, r_dest, label, address);
}

t_axe_instruction * gen_store_instruction
      (t_program_infos *program, int r_dest, t_axe_label *label, int address)
{
   return gen_unary_instruction(program, STORE, r_dest, label, address);
}

t_axe_instruction * gen_store_array_instruction
      (t_program_infos *program, int r_source1, int immediate, t_axe_label *label, int address)
{
   return gen_unary_immediate_jump_instruction(program, JIT_STORE_I, r_source1, immediate, label, 0);
}

t_axe_instruction * gen_load_array_instruction
      (t_program_infos *program, int r_source1, int immediate, t_axe_label *label, int address)
{
   return gen_unary_immediate_jump_instruction(program, JIT_LOAD_I, r_source1, immediate, label, 0);
}

t_axe_instruction * gen_jit_ldr_i_instruction
      (t_program_infos *program, int r_dest,int r_source1)
{
   return gen_binary_instruction(program, JIT_LDR_I, r_dest, r_source1, 0);
}

t_axe_instruction * gen_jit_sti_i_instruction
      (t_program_infos *program, int r_source1, int immediate, t_axe_label *label, int address)
{
   return gen_unary_immediate_jump_instruction(program, JIT_STI_I, r_source1, immediate, 0, 0);
}

/*-------------------------------------------------------
  *         UNARY IMMEDIATE  INSTRUCTIONS
  *------------------------------------------------------*/
/* MOVI Move immediate*/
t_axe_instruction * gen_jit_movi_i_instruction (t_program_infos *program
      , int r_dest, int immediate)
{
	return gen_unary_immediate_instruction (program, JIT_MOVI_I, r_dest, immediate);
}

 /*-------------------------------------------------------
 *              BINARY JUMP INSTRUCTIONS
 *------------------------------------------------------*/
/* BLTR	Branch lower than */
t_axe_instruction * gen_jit_bltr_i_instruction(t_program_infos *program
	, int r_source1, int r_source2, t_axe_label *label, int addr)
{
	gen_bin_jump_instruction (program, JIT_BLTR_I, r_source1, r_source2, label, addr);
}

/* BLER	Branch lower or equal */
t_axe_instruction * gen_jit_bler_i_instruction(t_program_infos *program
	, int r_source1, int r_source2, t_axe_label *label, int addr)
{
	gen_bin_jump_instruction (program, JIT_BLER_I, r_source1, r_source2, label, addr);
}

/* BGTR	Branch greater than */
t_axe_instruction * gen_jit_bgtr_i_instruction(t_program_infos *program
	, int r_source1, int r_source2, t_axe_label *label, int addr)
{
	gen_bin_jump_instruction (program, JIT_BGTR_I, r_source1, r_source2, label, addr);
}

/* BGER	Branch greater or equal */
t_axe_instruction * gen_jit_bger_i_instruction(t_program_infos *program
	, int r_source1, int r_source2, t_axe_label *label, int addr)
{
	gen_bin_jump_instruction (program, JIT_BGER_I, r_source1, r_source2, label, addr);
}

/* BEQR	Branch equal */
t_axe_instruction * gen_jit_beqr_i_instruction(t_program_infos *program
	, int r_source1, int r_source2, t_axe_label *label, int addr)
{
	gen_bin_jump_instruction (program, JIT_BEQR_I, r_source1, r_source2, label, addr);
}

/* BNER	Branch not equal */
t_axe_instruction * gen_jit_bner_i_instruction(t_program_infos *program
	, int r_source1, int r_source2, t_axe_label *label, int addr)
{
	gen_bin_jump_instruction (program, JIT_BNER_I, r_source1, r_source2, label, addr);
}

/* BMSR	Branch */
t_axe_instruction * gen_jit_bmsr_i_instruction(t_program_infos *program
	, int r_source1, int r_source2, t_axe_label *label, int addr)
{
	gen_bin_jump_instruction (program, JIT_BMSR_I, r_source1, r_source2, label, addr);
}

/* BMCR	Branch  */
t_axe_instruction * gen_jit_bmcr_i_instruction(t_program_infos *program
	, int r_source1, int r_source2, t_axe_label *label, int addr)
{
	gen_bin_jump_instruction (program, JIT_BMCR_I, r_source1, r_source2, label, addr);
}

/* BOADDR Branch on overflow after addition */
t_axe_instruction * gen_jit_boaddr_i_instruction(t_program_infos *program
	, int r_source1, int r_source2, t_axe_label *label, int addr)
{
	gen_bin_jump_instruction (program, JIT_BOADDR_I, r_source1, r_source2, label, addr);
}

/* BOSUBTR Branch on overflow after substraction */
t_axe_instruction * gen_jit_bosubr_i_instruction(t_program_infos *program
	, int r_source1, int r_source2, t_axe_label *label, int addr)
{
	gen_bin_jump_instruction (program, JIT_BOSUBR_I, r_source1, r_source2, label, addr);
}

 /*-------------------------------------------------------
 *           UNARY IMMEDIATE JUMP INSTRUCTIONS
 *------------------------------------------------------*/

/*BLTI Branch lower than immediate */
t_axe_instruction * gen_jit_blti_i_instruction(t_program_infos *program
	, int r_source1, int immediate, t_axe_label *label, int addr)
{
	gen_unary_immediate_jump_instruction(program, JIT_BLTI_I, r_source1, immediate, label, addr);
}

/*BLEI Branch lower or equal immediate */
t_axe_instruction * gen_jit_blei_i_instruction(t_program_infos *program
	, int r_source1, int immediate, t_axe_label *label, int addr)
{
	gen_unary_immediate_jump_instruction(program, JIT_BLEI_I, r_source1, immediate, label, addr);
}

/*BGTI Branch grater than immediate */
t_axe_instruction * gen_jit_bgti_i_instruction(t_program_infos *program
	, int r_source1, int immediate, t_axe_label *label, int addr)
{
	gen_unary_immediate_jump_instruction(program, JIT_BGTI_I, r_source1, immediate, label, addr);
}

/*BGEI Branch greater or equal immediate */
t_axe_instruction * gen_jit_bgei_i_instruction(t_program_infos *program
	, int r_source1, int immediate, t_axe_label *label, int addr)
{
	gen_unary_immediate_jump_instruction(program, JIT_BGEI_I, r_source1, immediate, label, addr);
}

/*BEQI Branch  equal immediate */
t_axe_instruction * gen_jit_beqi_i_instruction(t_program_infos *program
	, int r_source1, int immediate, t_axe_label *label, int addr)
{
	gen_unary_immediate_jump_instruction(program, JIT_BEQI_I, r_source1, immediate, label, addr);
}

/*BNEI Branch not equal immediate */
t_axe_instruction * gen_jit_bnei_i_instruction(t_program_infos *program
	, int r_source1, int immediate, t_axe_label *label, int addr)
{
	gen_unary_immediate_jump_instruction(program, JIT_BNEI_I, r_source1, immediate, label, addr);
}

/*BMSI Branch */
t_axe_instruction * gen_jit_bmsi_i_instruction(t_program_infos *program
	, int r_source1, int immediate, t_axe_label *label, int addr)
{
	gen_unary_immediate_jump_instruction(program, JIT_BMSI_I, r_source1, immediate, label, addr);
}

/*BMCI Branch  */
t_axe_instruction * gen_jit_bmci_i_instruction(t_program_infos *program
	, int r_source1, int immediate, t_axe_label *label, int addr)
{
	gen_unary_immediate_jump_instruction(program, JIT_BMCI_I, r_source1, immediate, label, addr);
}

/*BOADDI Branch on overflow after an addition  */
t_axe_instruction * gen_jit_boaddi_i_instruction(t_program_infos *program
	, int r_source1, int immediate, t_axe_label *label, int addr)
{
	gen_unary_immediate_jump_instruction(program, JIT_BOADDI_I, r_source1, immediate, label, addr);
}

/*BOSUBI Branch on overflow after a substraction  */
t_axe_instruction * gen_jit_bosubi_i_instruction(t_program_infos *program
	, int r_source1, int immediate, t_axe_label *label, int addr)
{
	gen_unary_immediate_jump_instruction(program, JIT_BOSUBI_I, r_source1, immediate, label, addr);
}

 /*-------------------------------------------------------
 *           UNCONDITIONAL JUMP INSTRUCTIONS
 *------------------------------------------------------*/

t_axe_instruction * gen_bt_instruction
      (t_program_infos *program, t_axe_label *label, int addr)
{
   return gen_jump_instruction (program, BT, label, addr);
}

t_axe_instruction * gen_bf_instruction
      (t_program_infos *program, t_axe_label *label, int addr)
{
   return gen_jump_instruction (program, BF, label, addr);
}

/* Generate halt instruction*/
t_axe_instruction * gen_halt_instruction
      (t_program_infos *program)
{
   t_axe_instruction *instr;

   /* test if program is initialized */
   if (program == NULL)
      notifyError(AXE_PROGRAM_NOT_INITIALIZED);
      
   /* create an instance of `t_axe_instruction' */
   instr = alloc_instruction(HALT);

   if (instr == NULL)
      notifyError(AXE_OUT_OF_MEMORY);

   /* add the newly created instruction to the current program */
   addInstruction(program, instr);

   /* return the load instruction */
   return instr;
}

/* Generate NOP instruction*/
t_axe_instruction * gen_nop_instruction(t_program_infos *program)
{
   t_axe_instruction *instr;

   /* test if program is initialized */
   if (program == NULL)
      notifyError(AXE_PROGRAM_NOT_INITIALIZED);
   
   /* create an instance of `t_axe_instruction' */
   instr = alloc_instruction(NOP);

   if (instr == NULL)
      notifyError(AXE_OUT_OF_MEMORY);

   /* add the newly created instruction to the current program */
   addInstruction(program, instr);

   /* return the load instruction */
   return instr;
}

/* Generate unary instructions*/
t_axe_instruction * gen_unary_instruction (t_program_infos *program
         , int opcode, int r_dest, t_axe_label *label, int addr)
{
   t_axe_instruction *instr;
   t_axe_register *reg;
   t_axe_address *address;
   int addressType;

   /* test if program is initialized */
   if (program == NULL)
      notifyError(AXE_PROGRAM_NOT_INITIALIZED);

   if (r_dest == REG_INVALID)
      notifyError(AXE_INVALID_REGISTER_INFO);
   
   /* test if value is correctly initialized */
   if (label != NULL)
   {
      if (label->labelID == LABEL_UNSPECIFIED)
      {
         notifyError(AXE_INVALID_LABEL);
      }

      /* address type is a label type */
      addressType = LABEL_TYPE;
   }
   else
   {
      if (addr < 0)
         notifyError(AXE_INVALID_ADDRESS);

      /* address type is a label type */
      addressType = ADDRESS_TYPE;

   }

   /* test if the opcode is a valid opcode */
   if (opcode == INVALID_OPCODE)
      notifyError(AXE_INVALID_OPCODE);

   /* create an instance of `t_axe_instruction' */
   instr = alloc_instruction(opcode);

   if (instr == NULL)
      notifyError(AXE_OUT_OF_MEMORY);

   /* initialize a register info */
   reg = alloc_register(r_dest, 0);

   if (reg == NULL)
   {
      free_Instruction(instr);
      notifyError(AXE_OUT_OF_MEMORY);
   }
   
   /* update the reg_1 info */
   instr->reg_1 = reg;

   /* initialize an address info */
   address = alloc_address(addressType, addr, label);
   
   if (address == NULL)
   {
      free_Instruction(instr);
      notifyError(AXE_OUT_OF_MEMORY);
   }

   /* update the instruction address info */
   instr->address = address;

   /* add the newly created instruction to the current program */
   addInstruction(program, instr);

   /* return the load instruction */
   return instr;
}

/*Generate binary instructions*/
t_axe_instruction * gen_binary_instruction (t_program_infos *program
      , int opcode, int r_dest, int r_source1, int immediate)
{
   t_axe_instruction *instr;
   t_axe_register *reg_dest;
   t_axe_register *reg_source1;

   /* test if program is initialized */
   if (program == NULL)
      notifyError(AXE_PROGRAM_NOT_INITIALIZED);

   /* test if value is correctly initialized */
   if (  (r_dest == REG_INVALID)
         || (r_source1 == REG_INVALID) )
   {
      notifyError(AXE_INVALID_REGISTER_INFO);
   }

   /* create an instance of `t_axe_instruction' */
   instr = alloc_instruction(opcode);

   if (instr == NULL)
      notifyError(AXE_OUT_OF_MEMORY);

   /* initialize a register info */
   reg_dest = alloc_register(r_dest, 0);
   if (reg_dest == NULL)
   {
      free_Instruction(instr);
      notifyError(AXE_OUT_OF_MEMORY);
   }

   /* update the reg_1 info */
   instr->reg_1 = reg_dest;

   reg_source1 = alloc_register(r_source1, 0);
   if (reg_source1 == NULL)
   {
      free_Instruction(instr);
      notifyError(AXE_OUT_OF_MEMORY);
      return NULL;
   }

   /* update the reg_1 info */
   instr->reg_2 = reg_source1;

   /* assign an immediate value */
   instr->immediate = immediate;

   /* add the newly created instruction to the current program */
   addInstruction(program, instr);

   /* return the load instruction */
   return instr;
}

/* Generate ternary instructions*/
t_axe_instruction * gen_ternary_instruction (t_program_infos *program
      , int opcode, int r_dest, int r_source1, int r_source2, int flags)
{
   t_axe_instruction *instr;
   t_axe_register *reg_dest;
   t_axe_register *reg_source1;
   t_axe_register *reg_source2;
   
   /* test if program is initialized */
   if (program == NULL)
      notifyError(AXE_PROGRAM_NOT_INITIALIZED);

   /* test if value is correctly initialized */
   if (  (r_dest == REG_INVALID)
         || (r_source1 == REG_INVALID)
         || (r_source2 == REG_INVALID))
   {
      notifyError(AXE_INVALID_REGISTER_INFO);
   }

   /* create an instance of `t_axe_instruction' */
   instr = alloc_instruction(opcode);

   if (instr == NULL)
      notifyError(AXE_OUT_OF_MEMORY);

   /* initialize a register info */
   reg_dest = alloc_register(r_dest, ((flags & CG_INDIRECT_DEST)? 1 : 0) );
   if (reg_dest == NULL)
   {
      free_Instruction(instr);
      notifyError(AXE_OUT_OF_MEMORY);
   }

   /* update the reg_1 info */
   instr->reg_1 = reg_dest;

   reg_source1 = alloc_register(r_source1, 0);
   if (reg_source1 == NULL)
   {
      free_Instruction(instr);
      notifyError(AXE_OUT_OF_MEMORY);
   }

   /* update the reg_2 info */
   instr->reg_2 = reg_source1;

   reg_source2 = alloc_register(r_source2, ((flags & CG_INDIRECT_SOURCE)? 1 : 0));
   if (reg_source1 == NULL)
   {
      free_Instruction(instr);
      notifyError(AXE_OUT_OF_MEMORY);
   }

   /* update the reg_3 info */
   instr->reg_3 = reg_source2;

   /* add the newly created instruction to the current program */
   addInstruction(program, instr);

   /* return the load instruction */
   return instr;
}

/* GENERATE BINARY JUMP INSTRUCTION */
t_axe_instruction * gen_bin_jump_instruction (t_program_infos *program
		, int opcode, int r_source1, int r_source2, t_axe_label *label, int addr)
{
	t_axe_instruction *instr;
	t_axe_address * address;
	t_axe_register *reg_source1;
	t_axe_register *reg_source2;
	int addressType;

	/* test if program is initialized */
	if (program == NULL)
		notifyError(AXE_PROGRAM_NOT_INITIALIZED);

	/* test if value is correctly initialized */
	if ( (      r_source1 == REG_INVALID)
	        || (r_source2 == REG_INVALID))
	{
		notifyError(AXE_INVALID_REGISTER_INFO);
	}
	
	/* test if value is correctly initialized */
	if (label != NULL)
	{
		if (label->labelID == LABEL_UNSPECIFIED)
			notifyError(AXE_INVALID_LABEL);

		addressType = LABEL_TYPE;
	}
	else
	{
		if (addr < 0)
			notifyError(AXE_INVALID_ADDRESS);

		addressType = ADDRESS_TYPE;
	}

	/* test if the opcode is a valid opcode */
	if (opcode == INVALID_OPCODE)
		notifyError(AXE_INVALID_OPCODE);

	/* create an instance of `t_axe_instruction' */
	instr = alloc_instruction(opcode);

	if (instr == NULL)
		notifyError(AXE_OUT_OF_MEMORY);

	/* initialize an address info */
	address = alloc_address(addressType, addr, label);
   
	if (address == NULL)
	{
		free_Instruction(instr);
		notifyError(AXE_OUT_OF_MEMORY);
	}

	/* update the instruction address info */
	instr->address = address;
	
	reg_source1 = alloc_register(r_source1, 0);
	if (reg_source1 == NULL)
	{
		free_Instruction(instr);
		notifyError(AXE_OUT_OF_MEMORY);
	}

	/* update the reg_2 info */
	instr->reg_1 = reg_source1;

	reg_source2 = alloc_register(r_source2, 0);
	if (reg_source1 == NULL)
	{
		free_Instruction(instr);
		notifyError(AXE_OUT_OF_MEMORY);
	}

	/* update the reg_3 info */
	instr->reg_2 = reg_source2;

	/* add the newly created instruction to the current program */
	addInstruction(program, instr);

	/* return the load instruction */
	return instr;
}

/* GENERATE UNARY JUMP INSTRUCTION */
t_axe_instruction * gen_unary_immediate_jump_instruction (t_program_infos *program
		, int opcode, int r_source1,int immediate, t_axe_label *label, int addr)
{
	t_axe_instruction *instr;
	t_axe_address * address;
	t_axe_register *reg_source1;
	int addressType;

	/* test if program is initialized */
	if (program == NULL)
		notifyError(AXE_PROGRAM_NOT_INITIALIZED);

	/* test if value is correctly initialized */
	if ( r_source1 == REG_INVALID)
	{
		notifyError(AXE_INVALID_REGISTER_INFO);
	}
	
	/* test if value is correctly initialized */
	if (label != NULL)
	{
		if (label->labelID == LABEL_UNSPECIFIED)
			notifyError(AXE_INVALID_LABEL);

		addressType = LABEL_TYPE;
	}
	else
	{
		if (addr < 0)
			notifyError(AXE_INVALID_ADDRESS);

		addressType = ADDRESS_TYPE;
	}

	/* test if the opcode is a valid opcode */
	if (opcode == INVALID_OPCODE)
		notifyError(AXE_INVALID_OPCODE);

	/* create an instance of `t_axe_instruction' */
	instr = alloc_instruction(opcode);

	if (instr == NULL)
		notifyError(AXE_OUT_OF_MEMORY);

	/* initialize an address info */
	address = alloc_address(addressType, addr, label);
   
	if (address == NULL)
	{
		free_Instruction(instr);
		notifyError(AXE_OUT_OF_MEMORY);
	}

	/* update the instruction address info */
	instr->address = address;
	
	reg_source1 = alloc_register(r_source1, 0);
	if (reg_source1 == NULL)
	{
		free_Instruction(instr);
		notifyError(AXE_OUT_OF_MEMORY);
	}

	/* update the reg_2 info */
	instr->reg_1 = reg_source1;

	instr->immediate=immediate;

	/* add the newly created instruction to the current program */
	addInstruction(program, instr);

	/* return the load instruction */
	return instr;
}




t_axe_instruction * gen_jump_instruction (t_program_infos *program
      , int opcode, t_axe_label *label, int addr)
{
   t_axe_instruction *instr;
   t_axe_address * address;
   int addressType;

   /* test if program is initialized */
   if (program == NULL)
      notifyError(AXE_PROGRAM_NOT_INITIALIZED);

   /* test if value is correctly initialized */
   if (label != NULL)
   {
      if (label->labelID == LABEL_UNSPECIFIED)
         notifyError(AXE_INVALID_LABEL);

      addressType = LABEL_TYPE;
   }
   else
   {
      if (addr < 0)
         notifyError(AXE_INVALID_ADDRESS);

      addressType = ADDRESS_TYPE;
   }

   /* test if the opcode is a valid opcode */
   if (opcode == INVALID_OPCODE)
      notifyError(AXE_INVALID_OPCODE);

   /* create an instance of `t_axe_instruction' */
   instr = alloc_instruction(opcode);

   if (instr == NULL)
      notifyError(AXE_OUT_OF_MEMORY);

   /* initialize an address info */
   address = alloc_address(addressType, addr, label);
   
   if (address == NULL)
   {
      free_Instruction(instr);
      notifyError(AXE_OUT_OF_MEMORY);t_axe_instruction * gen_neg_instruction (t_program_infos *program
      , int r_dest, int r_source, int flags)
{
   return gen_ternary_instruction
         (program, SUB, r_dest, REG_0, r_source, flags);
}
   }

   /* update the instruction address info */
   instr->address = address;

   /* add the newly created instruction to the current program */
   addInstruction(program, instr);

   /* return the load instruction */
   return instr;
}

/* Generate Unary Immediate Instruction */
t_axe_instruction * gen_unary_immediate_instruction(t_program_infos *program
		, int opcode, int r_dest, int immediate)
{
	
	t_axe_instruction *instr;
	t_axe_register *reg_dest;

	/* test if program is initialized */
	if (program == NULL)
		notifyError(AXE_PROGRAM_NOT_INITIALIZED);

	/* test if value is correctly initialized */
	if (r_dest == REG_INVALID)
	{
		notifyError(AXE_INVALID_REGISTER_INFO);
	}
	
	/* test if the opcode is a valid opcode */
	if (opcode == INVALID_OPCODE)
		notifyError(AXE_INVALID_OPCODE);

	/* create an instance of `t_axe_instruction' */
	instr = alloc_instruction(opcode);

	if (instr == NULL)
		notifyError(AXE_OUT_OF_MEMORY);

	/* initialize a register info */
	reg_dest = alloc_register(r_dest, 0);
	
	if (reg_dest == NULL)
	{
		free_Instruction(instr);
		notifyError(AXE_OUT_OF_MEMORY);
	}

	/* update the reg_1 info */
	instr->reg_1 = reg_dest;

	/* assign an immediate value */
	instr->immediate = immediate;

	/* add the newly created instruction to the current program */
	addInstruction(program, instr);

	/* return the instruction */
	return instr;
}
/* Generate Load/Store Instructions*/
t_axe_instruction * gen_load_store_instruction (t_program_infos *program
		, int opcode, int r_dest, int r_source1,int immediate, t_axe_label *label, int addr)
{
	t_axe_instruction *instr;
	t_axe_address * address;
	t_axe_register *reg_dest;
	t_axe_register *reg_source1;
	int addressType;

	/* test if program is initialized */
	if (program == NULL)
		notifyError(AXE_PROGRAM_NOT_INITIALIZED);

	/* test if value is correctly initialized */
	if ( (      r_dest == REG_INVALID) )
	{
		notifyError(AXE_INVALID_REGISTER_INFO);
	}
	
	/* test if value is correctly initialized */
	if (label != NULL)
	{
		if (label->labelID == LABEL_UNSPECIFIED)
			notifyError(AXE_INVALID_LABEL);

		addressType = LABEL_TYPE;
	}
	else
	{
		if (addr < 0)
			notifyError(AXE_INVALID_ADDRESS);

		addressType = ADDRESS_TYPE;
	}

	/* test if the opcode is a valid opcode */
	if (opcode == INVALID_OPCODE)
		notifyError(AXE_INVALID_OPCODE);

	/* create an instance of `t_axe_instruction' */
	instr = alloc_instruction(opcode);

	if (instr == NULL)
		notifyError(AXE_OUT_OF_MEMORY);

	/* initialize an address info */
	address = alloc_address(addressType, addr, label);
   
	if (address == NULL)
	{
		free_Instruction(instr);
		notifyError(AXE_OUT_OF_MEMORY);
	}

	/* update the instruction address info */
	instr->address = address;
	
	reg_dest = alloc_register(r_dest, 0);
	if (reg_source1 == NULL)
	{
		free_Instruction(instr);
		notifyError(AXE_OUT_OF_MEMORY);
	}

	/* update the reg_2 info */
	instr->reg_1 = reg_dest;

	reg_source1 = alloc_register(r_source1, 0);
	if (reg_source1 == NULL)
	{
		free_Instruction(instr);
		notifyError(AXE_OUT_OF_MEMORY);
	}

	/* update the reg_3 info */
	instr->reg_2 = reg_source1;
	
	instr->immediate=immediate;

	/* add the newly created instruction to the current program */
	addInstruction(program, instr);

	/* return the load instruction */
	return instr;
}

/* Generate Unary Immediate Instruction */
t_axe_instruction * gen_unary_rw_instruction(t_program_infos *program
		, int opcode, int r_dest)
{
	
	t_axe_instruction *instr;
	t_axe_register *reg_dest;

	/* test if program is initialized */
	if (program == NULL)
		notifyError(AXE_PROGRAM_NOT_INITIALIZED);

	/* test if value is correctly initialized */
	if (r_dest == REG_INVALID)
	{
		notifyError(AXE_INVALID_REGISTER_INFO);
	}
	
	/* test if the opcode is a valid opcode */
	if (opcode == INVALID_OPCODE)
		notifyError(AXE_INVALID_OPCODE);

	/* create an instance of `t_axe_instruction' */
	instr = alloc_instruction(opcode);

	if (instr == NULL)
		notifyError(AXE_OUT_OF_MEMORY);

	/* initialize a register info */
	reg_dest = alloc_register(r_dest, 0);
	
	if (reg_dest == NULL)
	{
		free_Instruction(instr);
		notifyError(AXE_OUT_OF_MEMORY);
	}

	/* update the reg_1 info */
	instr->reg_1 = reg_dest;

	/* add the newly created instruction to the current program */
	addInstruction(program, instr);

	/* return the instruction */
	return instr;
}

